package dev.fitko.fitconnect.client.util;

import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.destination.DestinationService;
import dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import dev.fitko.fitconnect.api.domain.model.metadata.data.SubmissionSchema;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.services.Sender;
import dev.fitko.fitconnect.client.sender.EncryptedSubmissionBuilder;
import dev.fitko.fitconnect.client.sender.SubmissionBuilder;
import dev.fitko.fitconnect.client.sender.model.EncryptedSubmissionPayload;
import dev.fitko.fitconnect.client.sender.model.SubmissionPayload;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;

import java.net.URI;
import java.util.Set;
import java.util.UUID;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.containsString;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

class ValidDataGuardTest {

    ValidDataGuard underTest;
    Sender senderMock;

    @BeforeEach
    void setUp() {
        senderMock = mock(Sender.class);
        underTest = new ValidDataGuard(senderMock);
    }

    @Nested
    class SubmissionPayloadDataTests {

        @Test
        void testValidSubmissionPayload() {

            // Given
            final SubmissionSchema submissionSchema = new SubmissionSchema();
            submissionSchema.setMimeType(MimeType.APPLICATION_JSON);
            submissionSchema.setSchemaUri(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"));

            final DestinationService service = new DestinationService();
            service.setIdentifier("urn:de:fim:leika:leistung:99400048079000");
            service.setSubmissionSchemas(Set.of(submissionSchema));

            final Destination destination = new Destination();
            destination.setDestinationId(UUID.randomUUID());
            destination.setServices(Set.of(service));

            when(senderMock.getDestination(any())).thenReturn(destination);
            when(senderMock.validateJsonFormat(any(), eq(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json")))).thenReturn(ValidationResult.ok());

            final SubmissionPayload submissionPayload = SubmissionBuilder.Builder()
                    .withJsonData("\"test\": \"json\"",
                            URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"))
                    .withDestination(UUID.randomUUID())
                    .withServiceType("Test", "urn:de:fim:leika:leistung:99400048079000")
                    .build();

            // Then valid if no exception is thrown
            underTest.ensureValidDataPayload(submissionPayload);
        }

        @Test
        void testMissingJsonData() {

            // Given
            final SubmissionPayload submissionPayload = SubmissionBuilder.Builder()
                    .withJsonData(null, URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"))
                    .withDestination(UUID.randomUUID())
                    .withServiceType("Test", "urn:de:fim:leika:leistung:99400048079000")
                    .build();

            // When
            final IllegalArgumentException exception = assertThrows(IllegalArgumentException.class, () -> underTest.ensureValidDataPayload(submissionPayload));

            // Then
            assertThat(exception.getMessage(), containsString("Data is mandatory, but was null"));
        }

        @Test
        void testMissingXmlData() {

            // Given
            final SubmissionPayload submissionPayload = SubmissionBuilder.Builder()
                    .withXmlData(null, URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"))
                    .withDestination(UUID.randomUUID())
                    .withServiceType("Test", "urn:de:fim:leika:leistung:99400048079000")
                    .build();

            // When
            final IllegalArgumentException exception = assertThrows(IllegalArgumentException.class, () -> underTest.ensureValidDataPayload(submissionPayload));

            // Then
            assertThat(exception.getMessage(), containsString("Data is mandatory, but was null"));
        }

        @Test
        void testMissingDestinationId() {

            // Given
            when(senderMock.validateJsonFormat(any(), eq(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json")))).thenReturn(ValidationResult.ok());

            final SubmissionPayload submissionPayload = SubmissionBuilder.Builder()
                    .withJsonData("{}", URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"))
                    .withDestination(null)
                    .withServiceType("name", "urn:de:fim:leika:leistung:99400048079000")
                    .build();

            // When
            final IllegalArgumentException exception = assertThrows(IllegalArgumentException.class, () -> underTest.ensureValidDataPayload(submissionPayload));

            // Then
            assertThat(exception.getMessage(), containsString("DestinationId is mandatory, but was null"));
        }

        @Test
        void testInvalidServiceIdentifier() {

            // Given
            when(senderMock.validateJsonFormat(any(), eq(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json")))).thenReturn(ValidationResult.ok());

            final SubmissionPayload submissionPayload = SubmissionBuilder.Builder()
                    .withJsonData("{\"test\" . \"data\"}",
                            URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"))
                    .withDestination(UUID.randomUUID())
                    .withServiceType("name", "illegal:test:identifier")
                    .build();
            // When
            final IllegalArgumentException exception = assertThrows(IllegalArgumentException.class, () -> underTest.ensureValidDataPayload(submissionPayload));

            // Then
            assertThat(exception.getMessage(), containsString("LeikaKey has invalid format"));
        }

        @Test
        void testMissingServiceIdentifier() {

            // Given
            when(senderMock.validateJsonFormat(any(), eq(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json")))).thenReturn(ValidationResult.ok());

            final SubmissionPayload submissionPayload = SubmissionBuilder.Builder()
                    .withJsonData("{\"test\" . \"data\"}",
                            URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"))
                    .withDestination(UUID.randomUUID())
                    .withServiceType("name", null)
                    .build();
            // When
            final IllegalArgumentException exception = assertThrows(IllegalArgumentException.class, () -> underTest.ensureValidDataPayload(submissionPayload));

            // Then
            assertThat(exception.getMessage(), containsString("Leika key is mandatory, but was null"));
        }

        @Test
        void testServiceIdentifierNotMatchingDestination() {

            // Given
            final DestinationService service = new DestinationService();
            service.setIdentifier("urn:de:fim:leika:leistung:99400048079000");

            final Destination destination = new Destination();
            destination.setServices(Set.of(service));

            when(senderMock.getDestination(any())).thenReturn(destination);
            when(senderMock.validateJsonFormat(any(), eq(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json")))).thenReturn(ValidationResult.ok());

            final SubmissionPayload submissionPayload = SubmissionBuilder.Builder()
                    .withJsonData("\"test\": \"json\"",
                            URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"))
                    .withDestination(UUID.randomUUID())
                    .withServiceType("Test", "urn:de:fim:leika:leistung:123456789101114")
                    .build();
            // When
            final IllegalArgumentException exception = assertThrows(IllegalArgumentException.class, () -> underTest.ensureValidDataPayload(submissionPayload));

            // Then
            assertThat(exception.getMessage(), containsString("Provided service type 'urn:de:fim:leika:leistung:123456789101114' is not allowed by the destination"));
        }

        @Test
        void testMimeTypeNotMatchingDestination() {

            // Given
            final SubmissionSchema submissionSchema = new SubmissionSchema();
            submissionSchema.setMimeType(MimeType.APPLICATION_XML);
            submissionSchema.setSchemaUri(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"));

            final DestinationService service = new DestinationService();
            service.setIdentifier("urn:de:fim:leika:leistung:99400048079000");
            service.setSubmissionSchemas(Set.of(submissionSchema));

            final Destination destination = new Destination();
            destination.setServices(Set.of(service));

            when(senderMock.getDestination(any())).thenReturn(destination);
            when(senderMock.validateJsonFormat(any(), eq(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json")))).thenReturn(ValidationResult.ok());

            final SubmissionPayload submissionPayload = SubmissionBuilder.Builder()
                    .withJsonData("\"test\": \"json\"",
                            URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"))
                    .withDestination(UUID.randomUUID())
                    .withServiceType("Test", "urn:de:fim:leika:leistung:99400048079000")
                    .build();
            // When
            final IllegalArgumentException exception = assertThrows(IllegalArgumentException.class, () -> underTest.ensureValidDataPayload(submissionPayload));

            // Then
            assertThat(exception.getMessage(), containsString("Combination of provided MIME type 'application/json' and schema URI " +
                    "'https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json' is not allowed by the destination"));
        }

        @Test
        void testSchemaUriNotMatchingDestination() {

            // Given
            final SubmissionSchema submissionSchema = new SubmissionSchema();
            submissionSchema.setMimeType(MimeType.APPLICATION_JSON);
            submissionSchema.setSchemaUri(URI.create("https://schema.fitko.de/fim/s00000000010_1.0.0.schema.json"));

            final DestinationService service = new DestinationService();
            service.setIdentifier("urn:de:fim:leika:leistung:99400048079000");
            service.setSubmissionSchemas(Set.of(submissionSchema));

            final Destination destination = new Destination();
            destination.setServices(Set.of(service));

            when(senderMock.getDestination(any())).thenReturn(destination);
            when(senderMock.validateJsonFormat(any(), eq(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json")))).thenReturn(ValidationResult.ok());

            final SubmissionPayload submissionPayload = SubmissionBuilder.Builder()
                    .withJsonData("\"test\": \"json\"",
                            URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"))
                    .withDestination(UUID.randomUUID())
                    .withServiceType("Test", "urn:de:fim:leika:leistung:99400048079000")
                    .build();
            // When
            final IllegalArgumentException exception = assertThrows(IllegalArgumentException.class, () -> underTest.ensureValidDataPayload(submissionPayload));

            // Then
            assertThat(exception.getMessage(), containsString("Combination of provided MIME type 'application/json' and schema URI " +
                    "'https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json' is not allowed by the destination"));
        }
    }

    @Nested
    class EncryptedSubmissionPayloadDataTests {

        @Test
        void testValidEncryptedSubmissionPayload() {

            // Given
            final DestinationService service = new DestinationService();
            service.setIdentifier("urn:de:fim:leika:leistung:99400048079000");

            final Destination destination = new Destination();
            destination.setDestinationId(UUID.randomUUID());
            destination.setServices(Set.of(service));

            when(senderMock.getDestination(any())).thenReturn(destination);
            when(senderMock.validateJsonFormat(any(), eq(URI.create("urn:something")))).thenReturn(ValidationResult.ok());

            final EncryptedSubmissionPayload encryptedSubmissionPayload = EncryptedSubmissionBuilder.Builder()
                    .withEncryptedData("4Y0sJhadfrQnNZXeS7Pqh73FvtF")
                    .withEncryptedMetadata("4Y0sJhadfrQnNZXeS7Pqh73FvtF4Y0sJhadfrQnNZXeS7Pqh73FvtF")
                    .withDestination(UUID.randomUUID())
                    .withServiceType("Test", "urn:de:fim:leika:leistung:99400048079000")
                    .build();

            // Then valid if no exception is thrown
            underTest.ensureValidDataPayload(encryptedSubmissionPayload);
        }

        @Test
        void testMissingEncryptedData() {

            // Given
            final EncryptedSubmissionPayload encryptedSubmissionPayload = EncryptedSubmissionBuilder.Builder()
                    .withEncryptedData(null)
                    .withEncryptedMetadata("4Y0sJhadfrQnNZXeS7Pqh73FvtF4Y0sJhadfrQnNZXeS7Pqh73FvtF")
                    .withDestination(UUID.randomUUID())
                    .withServiceType("Test", "urn:de:fim:leika:leistung:99400048079000")
                    .build();

            // When
            final IllegalArgumentException exception = assertThrows(IllegalArgumentException.class, () -> underTest.ensureValidDataPayload(encryptedSubmissionPayload));

            // Then
            assertThat(exception.getMessage(), containsString("Encrypted data is mandatory, but was null"));
        }

        @Test
        void testMissingEncryptedMetadata() {

            // Given
            final EncryptedSubmissionPayload encryptedSubmissionPayload = EncryptedSubmissionBuilder.Builder()
                    .withEncryptedData("4Y0sJhadfrQnNZXeS7Pqh73FvtF")
                    .withEncryptedMetadata(null)
                    .withDestination(UUID.randomUUID())
                    .withServiceType("Test", "urn:de:fim:leika:leistung:99400048079000")
                    .build();

            // When
            final IllegalArgumentException exception = assertThrows(IllegalArgumentException.class, () -> underTest.ensureValidDataPayload(encryptedSubmissionPayload));

            // Then
            assertThat(exception.getMessage(), containsString("Encrypted metadata must not be null"));
        }

        @Test
        void testMissingDestinationId() {

            // Given
            when(senderMock.validateJsonFormat(any(), eq(URI.create("urn:something")))).thenReturn(ValidationResult.ok());

            final EncryptedSubmissionPayload encryptedSubmissionPayload = EncryptedSubmissionBuilder.Builder()
                    .withEncryptedData("4Y0sJhadfrQnNZXeS7Pqh73FvtF")
                    .withEncryptedMetadata("4Y0sJhadfrQnNZXeS7Pqh73FvtF4Y0sJhadfrQnNZXeS7Pqh73FvtF")
                    .withDestination(null)
                    .withServiceType("name", "urn:de:fim:leika:leistung:99400048079000")
                    .build();

            // When
            final IllegalArgumentException exception = assertThrows(IllegalArgumentException.class, () -> underTest.ensureValidDataPayload(encryptedSubmissionPayload));

            // Then
            assertThat(exception.getMessage(), containsString("DestinationId is mandatory, but was null"));
        }

        @Test
        void testInvalidServiceIdentifier() {

            // Given
            when(senderMock.validateJsonFormat(any(), eq(URI.create("urn:something")))).thenReturn(ValidationResult.ok());

            final EncryptedSubmissionPayload encryptedSubmissionPayload = EncryptedSubmissionBuilder.Builder()
                    .withEncryptedData("4Y0sJhadfrQnNZXeS7Pqh73FvtF")
                    .withEncryptedMetadata("4Y0sJhadfrQnNZXeS7Pqh73FvtF4Y0sJhadfrQnNZXeS7Pqh73FvtF")
                    .withDestination(UUID.randomUUID())
                    .withServiceType("name", "illegal:test:identifier")
                    .build();

            // When
            final IllegalArgumentException exception = assertThrows(IllegalArgumentException.class, () -> underTest.ensureValidDataPayload(encryptedSubmissionPayload));

            // Then
            assertThat(exception.getMessage(), containsString("LeikaKey has invalid format"));
        }

        @Test
        void testMissingServiceIdentifier() {

            // Given
            when(senderMock.validateJsonFormat(any(), eq(URI.create("urn:something")))).thenReturn(ValidationResult.ok());

            final EncryptedSubmissionPayload encryptedSubmissionPayload = EncryptedSubmissionBuilder.Builder()
                    .withEncryptedData("4Y0sJhadfrQnNZXeS7Pqh73FvtF")
                    .withEncryptedMetadata("4Y0sJhadfrQnNZXeS7Pqh73FvtF4Y0sJhadfrQnNZXeS7Pqh73FvtF")
                    .withDestination(UUID.randomUUID())
                    .withServiceType("name", null)
                    .build();
            // When
            final IllegalArgumentException exception = assertThrows(IllegalArgumentException.class, () -> underTest.ensureValidDataPayload(encryptedSubmissionPayload));

            // Then
            assertThat(exception.getMessage(), containsString("Leika key is mandatory, but was null"));
        }

        @Test
        void testServiceIdentifierNotMatchingDestination() {

            // Given
            final DestinationService service = new DestinationService();
            service.setIdentifier("urn:de:fim:leika:leistung:99400048079000");

            final Destination destination = new Destination();
            destination.setServices(Set.of(service));

            when(senderMock.getDestination(any())).thenReturn(destination);
            when(senderMock.validateJsonFormat(any(), eq(URI.create("urn:something")))).thenReturn(ValidationResult.ok());

            final EncryptedSubmissionPayload encryptedSubmissionPayload = EncryptedSubmissionBuilder.Builder()
                    .withEncryptedData("4Y0sJhadfrQnNZXeS7Pqh73FvtF")
                    .withEncryptedMetadata("4Y0sJhadfrQnNZXeS7Pqh73FvtF4Y0sJhadfrQnNZXeS7Pqh73FvtF")
                    .withDestination(UUID.randomUUID())
                    .withServiceType("name", "urn:de:fim:leika:leistung:11111111111111")
                    .build();

            // When
            final IllegalArgumentException exception = assertThrows(IllegalArgumentException.class, () -> underTest.ensureValidDataPayload(encryptedSubmissionPayload));

            // Then
            assertThat(exception.getMessage(), containsString("Provided service type 'urn:de:fim:leika:leistung:11111111111111' is not allowed by the destination"));
        }
    }

    @Nested
    class MiscTests {

        @Test
        void testInvalidDestinationId() {

            // Given
            final var invalidUUID = "1234:132434:5678";

            // When
            final IllegalArgumentException exception = assertThrows(IllegalArgumentException.class, () ->
                    SubmissionBuilder.Builder()
                            .withJsonData("{}", URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"))
                            .withDestination(UUID.fromString(invalidUUID))
                            .withServiceType("name", "test:key")
                            .build());

            // Then
            assertThat(exception.getMessage(), containsString("Invalid UUID string: " + invalidUUID));
        }
    }
}